function M = emgaussian( X, K, varargin )
%emgaussian EM for a mixture of K gaussians
%
%INPUTS:
%  X -- N x d array of data points (no missing data)
%  K -- number of mixture components
%  optional arguments
%
%Outputs:
%  M.p -- 1 x K mixing proportions
%  M.mu -- K x d means
%  M.Sigma -- d x d x K covariances
%  M.post -- N x K posterior probabilities
%  M.ll -- log-likelihood
%
%Sergey Kirshner (skirshne@purdue.edu)
%March 20, 2012

% Default arguments
rstream = RandStream.create('mt19937ar', 'Seed', 0);
verbose = false;
random_restarts = 10;
kmeans_restarts = 10;
epsilon = 1e-05;

% Reading in optional arguments
opt_nargin = length(varargin);
if mod(opt_nargin,2) == 1
  error('Optional arguments are expected in pairs');
end

for i = 1:2:opt_nargin
  switch varargin{i}
    case 'stream' % Random number stream
      rstream=varargin{i+1};
    case 'verbose' % Displays intermediate outputs
      if varargin{i+1} == 1
        verbose = true;
      end
    case 'random_restarts' % Number of random restarts
      random_restarts = varargin{i+1};
    case 'kmeans_restarts' % Number of k-means restarts
      kmeans_restarts = varargin{i+1};
    case 'em_epsilon' % Threshold for stoppiing iterations
      epsilon = varargin{i+1};
    otherwise
  end
end

total_restarts = random_restarts + kmeans_restarts;

% May need to use global random number stream as Matlab Stat toolbox uses it
RandStream.setDefaultStream( rstream );

[N, d] = size(X); % dimensions

best_ll = -Inf; % best log-likelihood so far

for r = 1:total_restarts
  % Initializing
  if r <= kmeans_restarts
    % Using k-means initialization
    idx = kmeans( X, K );
    % Converting the vector into a matrix of memberships probabilities
    post = zeros([N d]);
    for c = 1:K
      post(idx==c,c) = 1;
    end
  else
    % Using random initialization
    post = gamrnd(1,1,N,K);
    for i = 1:N
      post(i,:) = post(i,:) / sum(post(i,:));
    end
  end
    
  ll = -Inf; % Not computing the initial log-likelihood as parameters are not yet set
  done = false;
  iter = 0;
  
  Sigma = zeros([d d K]); % allocating covariance matrices
  
  while ~done
    iter=iter+1;
    ll_old=ll;
    
    % M-step: Updating the parameters
    % Unnormalized mixture proportions
    p = sum(post, 1);
    % Means
    mu = post' * X;
    for c = 1:K
      mu(c,:)=mu(c,:)/p(c);
    end
    % Covariances
    for c = 1:K
      Sigma(:,:,c) = (X-repmat(mu(c,:),[N 1]))' * sparse(diag(post(:,c))) * (X-repmat(mu(c,:),[N 1]))/p(c);
    end
    % Normalizing p
    p = p / sum(p);
    
    % E-step: updating membership probability
    % Computing probability of each element belonging to each component
    for c = 1:K
      post(:,c) = mvnpdf( X, mu(c,:), Sigma(:,:,c) );
    end
    % Multiplying by the mixture probabilities
    post = post*diag(p);
    % Comput the log-likelihood
    ll = sum(log(sum(post,2)));
    % Normalizing membership probabilities -- not the most efficient but will do
    for i = 1:N
      post(i,:) = post(i,:) / sum(post(i,:));
    end
        
    % Decinding whether to stop
    if (ll-ll_old) / (N*d) < epsilon;
      % Change in log-likelihood is small
      done=true;
    end
    if verbose
      fprintf('Restart=%d, iteration=%d: log-likelihood=%0.6f\n', r, iter, ll/N );
    end
  end
  
  % Deciding whether the parameters learned so far are the best
  if ll > best_ll
    best_ll = ll;
    M.p = p;
    M.mu = mu;
    M.Sigma = Sigma;
    M.post = post;
    M.ll = ll;
  end
end

end



